
#include <drivers/spi.h>
#include <drivers/flash.h>

#define FLASH_DEVICE_NAME       "SPI_2"


#define FLASH_ON        (2)     //FLASH ON chip
#define FLASH_CS        (5)     //FLASH ON chip
#define FLASH_DI        (26)    //Flash DI
#define FLASH_CLK       (27)    //Flash Clock

/***************************** GPIO and LEDs ***************************************/
#define GPIO0_LABEL DT_PROP(DT_NODELABEL(gpio0), label)
#define GPIO0_STATUS DT_PROP(DT_NODELABEL(gpio0), status)

#define FLASH_ERASE_CMD (0xc7)
#define FLASH_JEDEC_ID_CMD (0x9f)
#define FLASH_WRITE_ENABLE_CMD (0x06)
#define FLASH_PAGE_PROGRAM_CMD (0x02)
#define FLASH_READ_DATA_CMD (0x03)
#define FLASH_STATUS_REG1 (0x05)
#define FLASH_STATUS_REG2 (0x35)
#define FLASH_BUSY_BIT (0x01)
#define FLASH_WRE_BIT (0x02)

#define FLASH_MAX_BLOCK_ADDR (0x0f)
#define FLASH_MAX_SECTOR_ADDR (0xff)
#define FLASH_MAX_PAGE_ADDR (0xff)

#define PAGE_SIZE (256)

const struct device *spi_dev;


// Data buffer to be used for moving data to and from the SPI
// 4 is added for 1 byte of Instruction + 3 bytes of Address
uint8_t data_buffer[PAGE_SIZE + 4] = {0};

struct spi_cs_control spi_cs = {
    .gpio_pin = FLASH_CS,
    .gpio_dt_flags = GPIO_ACTIVE_LOW,
    .delay = 0,
    .gpio_dev = NULL,
};

static const struct spi_config spi_cfg = {
    // SPI mode 0 (CPOL = 0, CPHA = 0) , and SPI mode 3 (CPOL = 1, CPHA = 1)
    // Use Mode 0, since in Mode-3 (CS)
    .operation = SPI_WORD_SET(8) | SPI_TRANSFER_MSB | SPI_MODE_CPOL | SPI_MODE_CPHA,
    //    .operation = SPI_WORD_SET(8) | SPI_TRANSFER_MSB,

    //	.operation = SPI_WORD_SET(8),
    .frequency = 4000000U,
    //	.frequency = 1000000U,
    .slave = 0,
    .cs = &spi_cs,
};

void spi_init(void)
{
//        const char *const spiName = "SPI_2";
        spi_dev = device_get_binding(FLASH_DEVICE_NAME);

        // To control the CS line
        spi_cs.gpio_dev = device_get_binding(GPIO0_LABEL);

        if (spi_dev == NULL)
        {
                printk("Could not get %s device\n", FLASH_DEVICE_NAME);
                return;
        }
}

int spi_send_recv(uint8_t *tx_buffer, uint16_t tx_len, uint8_t *rx_buffer, uint16_t rx_len)
{
        int err;

        const struct spi_buf tx_buf = {
            .buf = tx_buffer,
            .len = tx_len};
        const struct spi_buf_set tx = {
            .buffers = &tx_buf,
            .count = 1};

        struct spi_buf rx_buf = {
            .buf = rx_buffer,
            .len = rx_len,
        };
        const struct spi_buf_set rx = {
            .buffers = &rx_buf,
            .count = 1};

        err = spi_transceive(spi_dev, &spi_cfg, &tx, &rx);
        if (err)
        {
                printk("SPI error: %d\n", err);
        }
        return err;
}

int spi_flash_get_status_register(uint8_t reg, uint8_t *value)
{
        uint8_t tx_status;
        uint8_t rx_status[2];
        int err;
        tx_status = reg;
        err = spi_send_recv(&tx_status, 1, rx_status, 2);
        if (err)
        {
                printk("Status reg (%x) failed\n", tx_status);
                return err;
        }
        *value = rx_status[1];
        return 0;
}

int spi_flash_is_busy(uint8_t *ret_status)
{
        uint8_t status;
        int err;
        status = 0;
        err = spi_flash_get_status_register(FLASH_STATUS_REG1, &status);
        if (err)
        {
                printk("Reg (%x) status not available \n", FLASH_STATUS_REG1);
                return err;
        }
        *ret_status = (status & FLASH_BUSY_BIT);
        return 0;
}

int spi_flash_wait_to_clear_busy_bit(uint16_t sleep_time, uint8_t retry)
{
        uint8_t status;
        int err;
        while (retry--)
        {
                err = spi_flash_is_busy(&status);
                if (err)
                {
                        printk("Error: Couldn't get spi flash busy status\n");
                        return err;
                }
                if (!status)
                {
                        return 0;
                }
                k_sleep(K_MSEC(sleep_time));
        }
        printk("Err: Giving up waiting for clearing busy bit\n");
        return -EBUSY;
}

int spi_flash_is_write_enable(uint8_t *ret_status)
{
        uint8_t status;
        int err;
        status = 0;
        err = spi_flash_get_status_register(FLASH_STATUS_REG1, &status);
        if (err)
        {
                printk("Err(WRE): Reg (%x) status not available \n", FLASH_STATUS_REG1);
                return err;
        }
        *ret_status = (status & FLASH_WRE_BIT);
        return 0;
}

int spi_flash_wait_for_write_enable(uint16_t sleep_time, uint8_t retry)
{
        uint8_t status;
        int err;
        while (retry--)
        {
                err = spi_flash_is_write_enable(&status);
                if (err)
                {
                        printk("Error: Couldn't get spi flash WRE status\n");
                        return err;
                }
                if (status)
                {
                        return 0;
                }
                k_sleep(K_MSEC(sleep_time));
        }
        printk("Err: Giving up waiting for WRE bit\n");
        return -EBUSY;
}

int spi_flash_erase_no_wait(void)
{
        uint8_t tx_buffer;
        int err;

        // Enable write
        err = spi_flash_write_enable_wait();
        if (err)
        {
                printk("Err: SPI Erase fail, couldn't enable write\n");
                return err;
        }
        tx_buffer = FLASH_ERASE_CMD;
        return spi_send_recv(&tx_buffer, 1, NULL, 0);
}

int spi_flash_erase_wait(void)
{
        int err;
        err = spi_flash_erase_no_wait();
        if (err)
        {
                printk("Erase chip failed..!!\n");
                return err;
        }

        // Chip erase time is 2-6 seconds as per datasheet, so doing 10 iteration 1 sec apart
        return spi_flash_wait_to_clear_busy_bit(1000, 10);
}

int spi_flash_write_enable_no_wait(void)
{
        uint8_t tx_buffer;
        tx_buffer = FLASH_WRITE_ENABLE_CMD;
        return spi_send_recv(&tx_buffer, 1, NULL, 0);
}

int spi_flash_write_enable_wait(void)
{
        int err;
        err = spi_flash_write_enable_no_wait();
        if (err)
        {
                printk("Err: Write Enable failed..!!\n");
                return err;
        }

        // Write enable may take time based on the previous instruciton executed
        // e.g. if previous instruction was Erase, it may take upto 6 secods
        //      if previous instruction was Page Program, it may take upto 3 mS
        //  This API is meant to be called after Page Program, so using 10 tries of 3 mS
        //  For Erase to be complete, use blocking erase API
        return spi_flash_wait_for_write_enable(3, 10);
}

int spi_flash_get_jedec_id(uint8_t *mf, uint8_t *mem_type, uint8_t *capacity)
{
        uint8_t tx_buffer, rx_buffer[4];
        int err;
        tx_buffer = FLASH_JEDEC_ID_CMD;
        err = spi_send_recv(&tx_buffer, 1, rx_buffer, 4);
        if (err)
        {
                printk("Err: JEDEC ID not obtained\n");
                return err;
        }
        *mf = rx_buffer[1];
        *mem_type = rx_buffer[2];
        *capacity = rx_buffer[3];
        return 0;
}

int spi_flash_write_data(uint8_t block_addr, uint8_t sector_addr, uint8_t page_addr, uint8_t *tx_buffer, uint16_t tx_len)
{
        int err;
        // Validate the block_addr and data length
        if (block_addr > FLASH_MAX_BLOCK_ADDR)
        {
                printk("Err: Invalid block_addr (%x), max allowed (%x)\n", block_addr, FLASH_MAX_BLOCK_ADDR);
                return -EINVAL;
        }
        if (tx_len > PAGE_SIZE)
        {
                printk("Err: Invalid tx_len (%x), max allowed (%x)\n", tx_len, PAGE_SIZE);
                return -EINVAL;
        }

        // Check and wait for busy bit if it's not clear
        err = spi_flash_wait_to_clear_busy_bit(3, 10);
        if (err)
        {
                printk("Err: Failed to write, chip is busy\n");
                return err;
        }

        // Enable and wait for Write Enable bit
        err = spi_flash_write_enable_wait();
        if (err)
        {
                printk("Err: Failed to write. WRE disabled.\n");
                return err;
        }

        // Fill the address, data and send over SPI
        memset(data_buffer, 0, PAGE_SIZE + 4);
        data_buffer[0] = FLASH_PAGE_PROGRAM_CMD;
        data_buffer[1] = block_addr;
        data_buffer[2] = sector_addr;
        data_buffer[3] = page_addr;
        memcpy(data_buffer + 4, tx_buffer, tx_len);

        return spi_send_recv(data_buffer, tx_len + 4, NULL, 0);
}

int spi_flash_read_data(uint8_t block_addr, uint8_t sector_addr, uint8_t page_addr, uint8_t *rx_buffer, uint16_t rx_len)
{
        int err;
        uint8_t raddr_buffer[4];
        // Validate the block_addr and data length
        if (block_addr > FLASH_MAX_BLOCK_ADDR)
        {
                printk("Err: Invalid block_addr (%x), max allowed (%x)\n", block_addr, FLASH_MAX_BLOCK_ADDR);
                return -EINVAL;
        }
        if (rx_len > PAGE_SIZE)
        {
                printk("Err: Invalid tx_len (%x), max allowed (%x)\n", rx_len, PAGE_SIZE);
                return -EINVAL;
        }

        // Check and wait for busy bit if it's not clear
        err = spi_flash_wait_to_clear_busy_bit(3, 10);
        if (err)
        {
                printk("Err: Failed to read, chip is busy\n");
                return err;
        }

        // Fill the address, data and send over SPI
        memset(data_buffer, 0, PAGE_SIZE + 4);
        raddr_buffer[0] = FLASH_READ_DATA_CMD;
        raddr_buffer[1] = block_addr;
        raddr_buffer[2] = sector_addr;
        raddr_buffer[3] = page_addr;

        err = spi_send_recv(raddr_buffer, 4, data_buffer, rx_len + 4);
        if (err)
        {
                printk("Err: Read SPI failed\n");
                return err;
        }

        memcpy(rx_buffer, data_buffer + 4, rx_len);
        return 0;
}

int spi_test(void)
{
        int err;

        // Get jedec ID
        uint8_t mf, mem_type, capacity, jedec_status, erase_status;
        jedec_status = true;
        err = spi_flash_get_jedec_id(&mf, &mem_type, &capacity);
        if (mf==0xEF && mem_type==0x40 && capacity == 0x14)
        {
                printk("TEST PASS: JEDEC ID: %x %x %x\n", mf, mem_type, capacity);
        }
        else
        {
                printk("TEST FAIL: JEDEC ID not obtained!\n");
                jedec_status = false;
        }
 //       return;
        //Erase chip and wait for the erase to complete
        erase_status = true;
        err=spi_flash_erase_wait();
        if (err) {
                printk("TEST FAIL: Erasing chip failed!\n");
                erase_status = false;
        }
        else
        {
                printk("TEST PASS: Chip Erase done!\n");
        }

        // Write all the pages and verify the content
        uint8_t tx_buffer[PAGE_SIZE], rx_buffer[PAGE_SIZE], compare_pass;
        uint16_t i, j, k;
        uint32_t total_write, total_read, total_verified;

        // Fill the write buffer
        for (i = 0; i < PAGE_SIZE; i++)
        {
                tx_buffer[i] = i;
        }

        total_write = 0;

        //Write Entire Flash
        for (i=0; i<=FLASH_MAX_BLOCK_ADDR; i++) {
                for (j=0; j<=FLASH_MAX_SECTOR_ADDR; j++) {
                        err = spi_flash_write_data(i, j, 0, tx_buffer, PAGE_SIZE);
                        if (err) {
                                printk("TEST FAIL: Write failed for 0x%02X%02X%02X\n", i, j, 0);
                        }
                        else {
                                total_write+=PAGE_SIZE;
                                printk("TEST INFO: 256 bytes written at: 0x%02X%02X%02X  Wr(%u)\n", i, j, 0, total_write);
                        }
                }
        }

        total_read = 0;
        total_verified = 0;
        // Read each page and compare the data
        for (i = 0; i <= FLASH_MAX_BLOCK_ADDR; i++)
        {
                for (j = 0; j <= FLASH_MAX_SECTOR_ADDR; j++)
                {
                        memset(rx_buffer, 0, PAGE_SIZE);
                        err = spi_flash_read_data(i, j, 0, rx_buffer, PAGE_SIZE);
                        if (err)
                        {
                                printk("TEST FAIL: Read failed for 0x%02X%02X%02X\n", i, j, 0);
                        }
                        else
                        {
                                total_read += PAGE_SIZE;
                                compare_pass = true;
                                // Compare the data
                                for (k = 0; k < PAGE_SIZE; k++)
                                {
                                        if (tx_buffer[k] != rx_buffer[k])
                                        {
                                                printk("TEST FAIL: Data Mismatch at- 0x%02X%02X%02X Rx(%x)!=Tx(%x)\n", i, j, k, rx_buffer[k], tx_buffer[k]);
                                                compare_pass = false;
                                        }
                                        else
                                        {
                                                total_verified++;
                                        }
                                }
                                if (compare_pass == true)
                                {
                                        printk("TEST INFO: Verified 256 bytes write at: 0x%02X%02X%02X  Read(%u) Verified(%u)\n", i, j, 0, total_read, total_verified);
                                }
                        }
                }
        }

        printk("\n\n\t**** Flash Chip Verification Result ****\n");
        if (total_write == 1048576 && total_read == 1048576 && total_verified == 1048576)
        {
                printk("\nFlash Write/Read verification: PASS\n");
                printk("\tTEST RESULT: Bytes Written=%u, Bytes Read=%u, Bytes Verified=%u\n", total_write, total_read, total_verified);
        }
        else
        {
                printk("\nFlash Write/Read verification: FAIL\n");
                printk("\tTEST RESULT: Bytes Written=%u, Bytes Read=%u, Bytes Verified=%u\n", total_write, total_read, total_verified);
        }

        if (jedec_status == true)
        {
                printk("\nJEDEC ID Read: PASS\n");
                printk("\tTEST RESULT: JEDEC ID (%02X %02X %02X)\n", mf, mem_type, capacity);
        }
        else
        {
                printk("\nJEDEC ID Read: FAIL\n");
        }

        if (erase_status == true)
        {
                printk("\nErase Chip: PASS\n");
        }
        else
        {
                printk("\nErase Chip: FAIL\n");
        }

        if (total_write == 1048576 && total_read == 1048576 && total_verified == 1048576 && jedec_status == true && erase_status == true) {
                return 0;
        }
        return -1;
}

void spi_test_flash_api(void)
{
#define FLASH_NAME "W25Q80DV"
#define FLASH_TEST_REGION_OFFSET (0xff000)
#define FLASH_SECTOR_SIZE (4096)
#define FLASH_DEVICE DT_LABEL(DT_INST(0, jedec_spi_nor))

        const uint8_t expected[] = {0x55, 0xaa, 0x66, 0x99};
        const size_t len = sizeof(expected);
        uint8_t buf[sizeof(expected)];
        const struct device *flash_dev;
        int rc;
        
        printk("\n" FLASH_NAME " SPI flash testing\n");

        printk("==========================\n");

        flash_dev = device_get_binding(FLASH_NAME);
        if (!flash_dev)
        {
                printk("SPI flash driver %s was not found!\n",
                       FLASH_NAME);
                return;
        }

#ifdef CONFIG_FLASH_PAGE_LAYOUT
        struct flash_pages_info pages_info;
        size_t page_count, chip_size;
        page_count = flash_get_page_count(flash_dev);
        (void)flash_get_page_info_by_idx(flash_dev, 0, &pages_info);
        chip_size = page_count * pages_info.size;
        printk("Using %s, chip size: %u bytes (page: %u)\n",
               FLASH_NAME, chip_size, pages_info.size);

#endif

        /* Write protection needs to be disabled before each write or
         * erase, since the flash component turns on write protection
         * automatically after completion of write and erase
         * operations.
         */
        printk("\nTest 1: Flash erase\n");

        rc = flash_erase(flash_dev, FLASH_TEST_REGION_OFFSET,
                         FLASH_SECTOR_SIZE);
        if (rc != 0)
        {
                printk("Flash erase failed! %d\n", rc);
        }
        else
        {
                printk("Flash erase succeeded!\n");
        }


        printk("\nTest 2: Flash write\n");
        printk("Attempting to write %zu bytes\n", len);
        rc = flash_write(flash_dev, FLASH_TEST_REGION_OFFSET, expected, len);
        if (rc != 0)
        {
                printk("Flash write failed! %d\n", rc);
                return;
        }

        memset(buf, 0, len);
        rc = flash_read(flash_dev, FLASH_TEST_REGION_OFFSET, buf, len);
        if (rc != 0)
        {
                printk("Flash read failed! %d\n", rc);
                return;
        }

        if (memcmp(expected, buf, len) == 0)
        {
                printk("Data read matches data written. Good!!\n");
        }
        else
        {
                const uint8_t *wp = expected;
                const uint8_t *rp = buf;
                const uint8_t *rpe = rp + len;

                printk("Data read does not match data written!!\n");
                while (rp < rpe)
                {
                        printk("%08x wrote %02x read %02x %s\n",
                               (uint32_t)(FLASH_TEST_REGION_OFFSET + (rp - buf)),
                               *wp, *rp, (*rp == *wp) ? "match" : "MISMATCH");
                        ++rp;
                        ++wp;
                }
        }
}
